import fs from 'fs'
import cron from 'node-cron'
import crypto from 'crypto'
import config from "./config.js"
import i18n from "./i18n.js"
import mailer from './mailer.js'

const _config = config.get()
let _cache = null
const _readLinks = []
const _dbFile = _config.vault.db.filename

const init = () => {
	if(fs.existsSync(_dbFile)) {
		_cache = JSON.parse(fs.readFileSync(_dbFile))
	} else {
		_cache = []
	}
	clean()
	cron.schedule('* * * * *', clean)
}
const clean = () => {
	let needSave = false
	for(let i = _cache.length - 1; i >= 0; i--) {
		if(calcTimeLeft(_cache[i]) <= 0) {
			_cache.splice(i, 1)
			needSave = true
		}
	}
	if(needSave === true) save()
	for(let i = _readLinks.length - 1; i >= 0; i--) {
		if(Math.round(_config.vault.otLinkID.timeout - ((new Date() - new Date(_readLinks[i].date)) / 60000)) <= 0) {
			_readLinks.splice(i, 1)
		}
	}
}
const save = () => {
	fs.writeFileSync(_dbFile, JSON.stringify(_cache))
}

const newRandomNumber = (min, max) => {
	return Math.floor(Math.random() * (max - min) + min)
}
const newRandomString = (length) => {
	const randomBytes = crypto.randomBytes(Math.ceil(length / 2))
	return randomBytes.toString('hex').slice(0, length)
}
const calcReadsLeft = (item) => {
	if(item.maxReads <= 0) return -1
	return item.maxReads - item.readCount
}
const calcTimeLeft = (item) => {
	return Math.round(item.expireTime - ((new Date() - new Date(item.date)) / 60000))
}
const encrypt = (data) => {
	const cipher = crypto.createCipheriv(_config.vault.crypto.method, _config.vault.crypto.key, _config.vault.crypto.iv)
	let ciphertext = cipher.update(data, 'utf8', 'base64')
	ciphertext += cipher.final('base64')
	let value = cipher.getAuthTag().toString('hex') + '=' + ciphertext
	return value
}
const decrypt = (data) => {
	if(!data || data === '') return ''
  const decipher = crypto.createDecipheriv(
    _config.vault.crypto.method, 
    _config.vault.crypto.key,
    _config.vault.crypto.iv
  )
	const splitIndex = data.indexOf('=')
	const tag = data.substring(0, splitIndex)
	const value = data.substring(splitIndex + 1)
	decipher.setAuthTag(Buffer.from(tag, 'hex'))
	let plaintext = decipher.update(value, 'base64', 'utf8')
	plaintext += decipher.final('utf8')
	return plaintext
}
const emailReadNotify = (data) => {
	const email = {
		to: data.emailNotify,
		subject: i18n.capitalize(i18n.getString('secret_read')),
		text: i18n.capitalize(i18n.getString('secret_read'))
	}
	if(data.emailNotifySubject) email.subject = email.subject + ': ' + data.emailNotifySubject
	else if(data.subject) email.subject = email.subject + ': ' + data.subject
	if(data.emailNotifyMessage) {
		if(data.message) {
			email.text = data.emailNotifyMessage
			email.text = email.text + '\n\n' + data.message
		} else {
			email.text = data.emailNotifyMessage
		}
	}	else if(data.message) email.text = data.message
	else email.text = email.subject
	mailer.send(email)
}

if(! _cache) init()

const createData = (data = {
	senderName: null,
	subject: null,
	message: null,
	secret: null,
	expireDays: null,
	expireHours: null,
	expireTime: null,
	maxReads: null,
	showMethod: null,
	showMethodEMail: null,
	emailNotify: null,
	emailNotifySubject: null,
	emailNotifyMessage: null,
}) => {
	if(! data) return null
	const validatedData = {}
	if(data.senderName && data.senderName != "") validatedData.senderName = data.senderName
	if(data.subject && data.subject != "") validatedData.subject = data.subject
	if(data.message && data.message != "") validatedData.message = data.message
	if(data.secret && data.secret != "") validatedData.secret = data.secret
	if(data.maxReads && data.maxReads != "") validatedData.maxReads = data.maxReads
	let expireTime = 0
	if(data.expireDays && data.expireDays > 0) expireTime = +data.expireDays * 24 * 60
	if(data.expireHours && data.expireHours > 0) expireTime = expireTime + +data.expireHours * 60
	if(data.expireTime && data.expireTime > 0) expireTime = +data.expireTime
	if(expireTime > 0) validatedData.expireTime = expireTime
	switch(data.showMethod) {
		case 'email':
			if(data.showMethodEMail && data.showMethodEMail != '') {
				validatedData.showMethod = data.showMethod
				validatedData.showMethodEMail = data.showMethodEMail
			} else {
				validatedData.showMethod = 'link'
			}
			break
		default:
			validatedData.showMethod = 'link'
	}
	if(data.showEMmail && data.showEMmail != "") validatedData.showEMmail = data.showEMmail
	if(data.emailNotify && data.emailNotify != "") validatedData.emailNotify = data.emailNotify
	if(data.emailNotifySubject && data.emailNotifySubject != "") validatedData.emailNotifySubject = data.emailNotifySubject
	if(data.emailNotifyMessage && data.emailNotifyMessage != "") validatedData.emailNotifyMessage = data.emailNotifyMessage

	if(! validatedData.secret || validatedData.secret === '') return null
	if(! validatedData.expireTime) validatedData.expireTime = 24 * 60
	else if(validatedData.expireTime < 1) validatedData.expireTime = 1
	else if(validatedData.expireTime > 560 * 60) validatedData.expireTime = 560 * 60
	if(! validatedData.maxReads || isNaN(validatedData.maxReads)) validatedData.maxReads = 1
	else if(validatedData.maxReads < 0) validatedData.maxReads = 0
	else if(validatedData.maxReads > 999) validatedData.maxReads = 999

	return validatedData
}
const createLinkData = (data) => {
	if(data) return {
		date: data.date,
		publicID: data.publicID,
		senderName: data.senderName,
		subject: data.subject,
		message: data.message,
		readsLeft: calcReadsLeft(data),
		timeLeft: calcTimeLeft(data),
	}
	return null
}
const create = (secretData) => {
	const data = createData(secretData)
	if(! data) return null
	data.date = new Date()
	data.readCount = 0
	data.publicID = newRandomString(newRandomNumber(_config.vault.publicID.minLength, _config.vault.publicID.maxLength))
	data.secret = encrypt(data.secret)
	_cache.push(data)
	save()
	const returnData = createLinkData(data)
	returnData.emailNotify = data.emailNotify
	returnData.emailOTL = data.emailOTL
	return returnData
}
const remove = (publicID) => {
	for(let i = 0; i < _cache.length; i++) {
		if(_cache[i].publicID == publicID) {
			_cache.splice(i, 1)
			save()
			for(let i = _readLinks.length - 1; i >= 0; i--) {
				if(_readLinks[i].publicID == publicID) {
					_readLinks.splice(i, 1)
				}
			}
			return true
		}
	}
	return false
}
const getLink = (publicID) => {
	const data = _cache.find((item) => item.publicID === publicID)
	return createLinkData(data)
}
const createOTLink = (publicID) => {
	const value = _cache.find((item) => item.publicID === publicID)
	if(value) {
		const readLink = {
			date: new Date(),
			otLinkID: newRandomString(newRandomNumber(_config.vault.otLinkID.minLength, _config.vault.otLinkID.maxLength)),
			publicID: publicID,
		}
		_readLinks.push(readLink)
		return {
			otLinkID: readLink.otLinkID
		}
	}
	return null
}
const removeOTL = (otLinkID) => {
	for(let i = _readLinks.length - 1; i >= 0; i--) {
		if(_readLinks[i].otLinkID == otLinkID) {
			_readLinks.splice(i, 1)
		}
	}
}
const read = (otLinkID) => {
	const linkValue = _readLinks.find((item) => item.otLinkID === otLinkID)
	if(linkValue) {
		removeOTL(otLinkID)
		const value = _cache.find((item) => item.publicID === linkValue.publicID)
		if(value) {
			value.readCount += 1
			if(value.readCount < value.maxReads) {
				save()
			} else {
				remove(value.publicID)
			}
			value.timeLeft = calcTimeLeft(value)
			value.readsLeft = calcReadsLeft(value)
			if(value.emailNotify) emailReadNotify(value)
			const returnValue = structuredClone(value)
			returnValue.secret = decrypt(returnValue.secret)
			return returnValue
		}
	}
	return null
}


export default {
	create,
	getLink,
	createOTLink,
	read,
}